@@ -33,6 +33,8 @@ from .urls import (
     SPREADSHEET_BATCH_UPDATE_URL,
     SPREADSHEET_VALUES_APPEND_URL,
     SPREADSHEET_VALUES_CLEAR_URL,
+    SPREADSHEET_DRIVE_URL,
+    WORKSHEET_DRIVE_URL,
     SPREADSHEET_VALUES_BATCH_UPDATE_URL
 )
 
@@ -93,6 +95,11 @@ class Spreadsheet(object):
             self._properties.update(metadata['properties'])
             return self._properties['title']
 
+    @property
+    def url(self):
+        """Spreadsheet URL"""
+        return SPREADSHEET_DRIVE_URL % self.id
+
     @property
     def updated(self):
         """.. deprecated:: 2.0
@@ -247,8 +254,9 @@ class Spreadsheet(object):
         r = self.client.request('get', url, params=params)
         return r.json()
 
-    def fetch_sheet_metadata(self):
-        params = {'includeGridData': 'false'}
+    def fetch_sheet_metadata(self, params=None):
+        if params is None:
+            params = {'includeGridData': 'false'}
 
         url = SPREADSHEET_URL % self.id
 
@@ -313,7 +321,7 @@ class Spreadsheet(object):
         except (StopIteration, KeyError):
             raise WorksheetNotFound(title)
 
-    def add_worksheet(self, title, rows, cols):
+    def add_worksheet(self, title, rows, cols, index=None):
         """Adds a new worksheet to a spreadsheet.
 
         :param title: A title of a new worksheet.
@@ -322,6 +330,8 @@ class Spreadsheet(object):
         :type rows: int
         :param cols: Number of columns.
         :type cols: int
+        :param index: Position of the sheet.
+        :type index: int
 
         :returns: a newly created :class:`worksheets <gspread.models.Worksheet>`.
         """
@@ -340,6 +350,9 @@ class Spreadsheet(object):
             }]
         }
 
+        if index is not None:
+            body["requests"][0]["addSheet"]["properties"]["index"] = index
+
         data = self.batch_update(body)
 
         properties = data['replies'][0]['addSheet']['properties']
@@ -402,12 +415,51 @@ class Spreadsheet(object):
         """
         body = {
             'requests': [{
-                'deleteSheet': {'sheetId': worksheet._properties['sheetId']}
+                'deleteSheet': {'sheetId': worksheet.id}
             }]
         }
 
         return self.batch_update(body)
 
+    def reorder_worksheets(self, worksheets_in_desired_order):
+        """Updates the ``index`` property of each Worksheets to reflect
+        its index in the provided sequence of Worksheets.
+
+        :param worksheets_in_desired_order: Iterable of Worksheet objects in desired order.
+
+        Note: If you omit some of the Spreadsheet's existing Worksheet objects from
+        the provided sequence, those Worksheets will be appended to the end of the sequence
+        in the order that they appear in the list returned by ``Spreadsheet.worksheets()``.
+
+        .. versionadded:: 3.4
+
+        """
+        idx_map = {}
+        for idx, w in enumerate(worksheets_in_desired_order):
+            idx_map[w.id] = idx
+        for w in self.worksheets():
+            if w.id in idx_map:
+                continue
+            idx += 1
+            idx_map[w.id] = idx
+
+        body = {
+            'requests': [
+                {
+                    'updateSheetProperties': {
+                        'properties': {
+                            'sheetId': key,
+                            'index': val
+                        },
+                        'fields': 'index'
+                    }
+                }
+                for key, val in idx_map.items()
+            ]
+        }
+
+        return self.batch_update(body)
+
     def share(self, value, perm_type, role, notify=True, email_message=None, with_link=False):
         """Share the spreadsheet with other accounts.
 
@@ -511,6 +563,11 @@ class Worksheet(object):
         """Worksheet title."""
         return self._properties['title']
 
+    @property
+    def url(self):
+        """Worksheet URL"""
+        return WORKSHEET_DRIVE_URL % (self.spreadsheet.id, self.id)
+
     @property
     def updated(self):
         """.. deprecated:: 2.0
@@ -535,6 +592,16 @@ class Worksheet(object):
         """Number of columns."""
         return self._properties['gridProperties']['columnCount']
 
+    @property
+    def frozen_row_count(self):
+        """Number of frozen rows."""
+        return self._properties['gridProperties'].get('frozenRowCount', 0)
+
+    @property
+    def frozen_col_count(self):
+        """Number of frozen columns."""
+        return self._properties['gridProperties'].get('frozenColumnCount', 0)
+
     def acell(self, label, value_render_option='FORMATTED_VALUE'):
         """Returns an instance of a :class:`gspread.models.Cell`.
 
@@ -582,14 +649,13 @@ class Worksheet(object):
 
         """
 
-        range_label = '%s!%s' % (self.title, rowcol_to_a1(row, col))
-        data = self.spreadsheet.values_get(
-            range_label,
-            params={'valueRenderOption': value_render_option}
-        )
-
         try:
-            value = data['values'][0][0]
+            data = self.get(
+                rowcol_to_a1(row, col),
+                value_render_option=value_render_option
+            )
+
+            value = data.first()
         except KeyError:
             value = ''
 
@@ -626,7 +692,7 @@ class Worksheet(object):
 
         """
 
-        range_label = '%s!%s' % (self.title, name)
+        range_label = absolute_range_name(self.title, name)
 
         data = self.spreadsheet.values_get(range_label)
 
@@ -648,15 +714,26 @@ class Worksheet(object):
             for j, value in enumerate(row)
         ]
 
-    def get_all_values(self):
+    def get_all_values(self, value_render_option='FORMATTED_VALUE'):
         """Returns a list of lists containing all cells' values as strings.
 
+        :param value_render_option: (optional) Determines how values should be
+                                    rendered in the the output. See
+                                    `ValueRenderOption`_ in the Sheets API.
+        :type value_render_option: str
+
+        .. _ValueRenderOption: https://developers.google.com/sheets/api/reference/rest/v4/ValueRenderOption
+
         .. note::
 
             Empty trailing rows and columns will not be included.
         """
+        range_name = absolute_range_name(self.title)
 
-        data = self.spreadsheet.values_get(self.title)
+        data = self.spreadsheet.values_get(
+            range_name,
+            params={'valueRenderOption': value_render_option}
+        )
 
         try:
             return fill_gaps(data['values'])
@@ -715,7 +792,12 @@ class Worksheet(object):
 
         return [dict(zip(keys, row)) for row in values]
 
-    def row_values(self, row, value_render_option='FORMATTED_VALUE'):
+    @accepted_kwargs(
+        major_dimension=None,
+        value_render_option=None,
+        date_time_render_option=None
+    )
+    def row_values(self, row, **kwargs):
         """Returns a list of all values in a `row`.
 
         Empty cells in this list will be rendered as :const:`None`.
@@ -730,16 +812,9 @@ class Worksheet(object):
         .. _ValueRenderOption: https://developers.google.com/sheets/api/reference/rest/v4/ValueRenderOption
 
         """
-
-        range_label = '%s!A%s:%s' % (self.title, row, row)
-
-        data = self.spreadsheet.values_get(
-            range_label,
-            params={'valueRenderOption': value_render_option}
-        )
-
         try:
-            return data['values'][0]
+            data = self.get('A{}:{}'.format(row, row), **kwargs)
+            return data[0]
         except KeyError:
             return []
 
@@ -760,10 +835,12 @@ class Worksheet(object):
         """
 
         start_label = rowcol_to_a1(1, col)
-        range_label = '%s!%s:%s' % (self.title, start_label, start_label[:-1])
+        range_label = '{}:{}'.format(start_label, start_label[:-1])
+
+        range_name = absolute_range_name(self.title, range_label)
 
         data = self.spreadsheet.values_get(
-            range_label,
+            range_name,
             params={
                 'valueRenderOption': value_render_option,
                 'majorDimension': 'COLUMNS'
@@ -804,10 +881,10 @@ class Worksheet(object):
             worksheet.update_cell(1, 1, '42')
 
         """
-        range_label = '%s!%s' % (self.title, rowcol_to_a1(row, col))
+        range_name = absolute_range_name(self.title, rowcol_to_a1(row, col))
 
         data = self.spreadsheet.values_update(
-            range_label,
+            range_name,
             params={
                 'valueInputOption': 'USER_ENTERED'
             },
@@ -847,10 +924,13 @@ class Worksheet(object):
         start = rowcol_to_a1(min(c.row for c in cell_list), min(c.col for c in cell_list))
         end = rowcol_to_a1(max(c.row for c in cell_list), max(c.col for c in cell_list))
 
-        range_label = '%s!%s:%s' % (self.title, start, end)
+        range_name = absolute_range_name(
+            self.title,
+            '{}:{}'.format(start, end)
+        )
 
         data = self.spreadsheet.values_update(
-            range_label,
+            range_name,
             params={
                 'valueInputOption': value_input_option
             },
@@ -901,10 +981,7 @@ class Worksheet(object):
         .. versionadded:: 3.3
 
         """
-        if range_name:
-            range_name = '%s!%s' % (self.title, range_name)
-        else:
-            range_name = self.title
+        range_name = absolute_range_name(self.title, range_name)
 
         params = filter_dict_values({
             'majorDimension': kwargs['major_dimension'],
@@ -947,7 +1024,7 @@ class Worksheet(object):
             worksheet.batch_get(['A1:B2', 'F12'])
 
         """
-        ranges = ['%s!%s' % (self.title, r) for r in ranges if r]
+        ranges = [absolute_range_name(self.title, r) for r in ranges if r]
 
         params = filter_dict_values({
             'majorDimension': kwargs['major_dimension'],
@@ -1018,10 +1095,10 @@ class Worksheet(object):
 
         """
         if is_scalar(range_name):
-            range_name = '%s!%s' % (self.title, range_name)
+            range_name = absolute_range_name(self.title, range_name)
         else:
             values = range_name
-            range_name = self.title
+            range_name = absolute_range_name(self.title)
 
         if is_scalar(values):
             values = [[values]]
@@ -1196,6 +1273,72 @@ class Worksheet(object):
 
         return self.spreadsheet.batch_update(body)
 
+    def sort(self, *specs, range=None):
+        """Sorts worksheet using given sort orders.
+
+        :param specs: The sort order per column. Each sort order represented
+        by a tuple where the 1st element is a column index and 2nd element is
+        the order itself: 'asc' or 'des'.
+        :type specs: List[Tuple[int, str]]
+        :param range: The range to sort. By default sorts whole sheet excluding
+        frozen rows. Range notation: 'A1:A1'.
+        :type range: str
+
+        Example::
+
+            # Sort sheet A -> Z by column 'B'
+            wks.sort((2, 'asc'))
+
+            # Sort range A2:G8 basing on column 'G' A -> Z and column 'B' Z -> A
+            wks.sort((7, 'asc'), (2, 'des'), range='A2:G8')
+
+        .. versionadded:: 3.4
+
+        """
+        if range:
+            start_a1, end_a1 = range.split(':')
+            start_row, start_col = a1_to_rowcol(start_a1)
+            end_row, end_col = a1_to_rowcol(end_a1)
+        else:
+            start_row = self._properties['gridProperties'].get('frozenRowCount', 0) + 1
+            start_col = 1
+            end_row = self.row_count
+            end_col = self.col_count
+
+        request_range = {
+            'sheetId': self.id,
+            'startRowIndex': start_row - 1,
+            'endRowIndex': end_row,
+            'startColumnIndex': start_col - 1,
+            'endColumnIndex': end_col
+        }
+
+        request_sort_specs = list()
+        for col, order in specs:
+            if order == 'asc':
+                request_order = 'ASCENDING'
+            elif order == 'des':
+                request_order = 'DESCENDING'
+            else:
+                raise ValueError("Either 'asc' or 'des' should be specified as sort order.")
+            request_sort_spec = {
+                "dimensionIndex": col - 1,
+                "sortOrder": request_order
+            }
+            request_sort_specs.append(request_sort_spec)
+
+        body = {
+            'requests': [{
+                "sortRange": {
+                    'range': request_range,
+                    'sortSpecs': request_sort_specs
+                }
+            }]
+        }
+
+        response = self.spreadsheet.batch_update(body)
+        return response
+
     def update_title(self, title):
         """Renames the worksheet.
 
@@ -1220,6 +1363,33 @@ class Worksheet(object):
         self._properties['title'] = title
         return response
 
+    def update_index(self, index):
+        """Updates the ``index`` property for the worksheet.
+
+        See the `Sheets API documentation
+        <https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets#sheetproperties>`_
+        for information on how updating the index property affects the order of worksheets
+        in a spreadsheet.
+
+        To reorder all worksheets in a spreadsheet, see `Spreadsheet.reorder_worksheets`.
+
+        .. versionadded:: 3.4
+
+        """
+        body = {
+            'requests': [{
+                'updateSheetProperties': {
+                    'properties': {
+                        'sheetId': self.id,
+                        'index': index
+                    },
+                    'fields': 'index'
+                }
+            }]
+        }
+
+        return self.spreadsheet.batch_update(body)
+
     def add_rows(self, rows):
         """Adds rows to worksheet.
 
@@ -1269,20 +1439,52 @@ class Worksheet(object):
         .. _InsertDataOption: https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets.values/append#InsertDataOption
 
         """
-
-        range_label = (
-            '%s!%s' % (self.title, table_range)
-            if table_range
-            else self.title
+        return self.append_rows(
+            [values],
+            value_input_option=value_input_option,
+            insert_data_option=insert_data_option,
+            table_range=table_range
         )
 
+    def append_rows(
+        self,
+        values,
+        value_input_option='RAW',
+        insert_data_option=None,
+        table_range=None
+    ):
+        """Adds multiple rows to the worksheet and populates them with values.
+        Widens the worksheet if there are more values than columns.
+
+        :param values: List of rows each row is List of values for the new row.
+        :param value_input_option: (optional) Determines how input data should
+                                    be interpreted. See `ValueInputOption`_ in
+                                    the Sheets API.
+        :type value_input_option: str
+        :param insert_data_option: (optional) Determines how the input data
+                                    should be inserted. See
+                                    `InsertDataOption`_ in the Sheets API
+                                    reference.
+        :type insert_data_option: str
+        :param table_range: (optional) The A1 notation of a range to search for
+                             a logical table of data. Values are appended after
+                             the last row of the table.
+                             Examples: `A1` or `B2:D4`
+        :type table_range: str
+
+        .. _ValueInputOption: https://developers.google.com/sheets/api/reference/rest/v4/ValueInputOption
+        .. _InsertDataOption: https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets.values/append#InsertDataOption
+
+        """
+        range_label = absolute_range_name(self.title, table_range)
+
         params = {
             'valueInputOption': value_input_option,
             'insertDataOption': insert_data_option
         }
 
         body = {
-            'values': [values]
+            'values': values
         }
 
         return self.spreadsheet.values_append(range_label, params, body)
@@ -1325,7 +1527,7 @@ class Worksheet(object):
 
         self.spreadsheet.batch_update(body)
 
-        range_label = '%s!%s' % (self.title, 'A%s' % index)
+        range_label = absolute_range_name(self.title, 'A%s' % index)
 
         data = self.spreadsheet.values_update(
             range_label,
@@ -1360,13 +1562,38 @@ class Worksheet(object):
 
         return self.spreadsheet.batch_update(body)
 
+    def delete_rows(self, start_index, end_index):
+        """"Deletes multi rows from the worksheet at the specified index.
+
+        :param start_index: Index of a first row for deletion.
+        :param end_index: Index of a last row for deletion.
+        :type index: int
+        """
+        body = {
+            "requests": [{
+                "deleteDimension": {
+                    "range": {
+                      "sheetId": self.id,
+                      "dimension": "ROWS",
+                      "startIndex": start_index - 1,
+                      "endIndex": end_index - 1
+                    }
+                }
+            }]
+        }
+
+        return self.spreadsheet.batch_update(body)
+
     def clear(self):
         """Clears all cells in the worksheet.
         """
-        return self.spreadsheet.values_clear(self.title)
+        return self.spreadsheet.values_clear(
+            absolute_range_name(self.title)
+        )
 
     def _finder(self, func, query):
-        data = self.spreadsheet.values_get(self.title)
+
+        data = self.spreadsheet.values_get(absolute_range_name(self.title))
 
         try:
             values = fill_gaps(data['values'])
@@ -1407,6 +1634,102 @@ class Worksheet(object):
         """
         return list(self._finder(filter, query))
 
+    def freeze(self, rows=None, cols=None):
+        """Freeze rows and/or columns on the worksheet.
+
+        :param rows: Number of rows to freeze.
+        :param cols: Number of columns to freeze.
+        """
+        grid_properties = {}
+
+        if rows is not None:
+            grid_properties['frozenRowCount'] = rows
+
+        if cols is not None:
+            grid_properties['frozenColumnCount'] = cols
+
+        if not grid_properties:
+            raise TypeError("Either 'rows' or 'cols' should be specified.")
+
+        fields = ','.join(
+            'gridProperties/%s' % p for p in grid_properties.keys()
+        )
+
+        body = {
+            'requests': [{
+                'updateSheetProperties': {
+                    'properties': {
+                        'sheetId': self.id,
+                        'gridProperties': grid_properties
+                    },
+                    'fields': fields
+                }
+            }]
+        }
+
+        return self.spreadsheet.batch_update(body)
+
+    @cast_to_a1_notation
+    def set_basic_filter(self, name=None):
+        """Add a basic filter to the worksheet. If a range or bundaries
+        are passed, the filter will be limited to the given range.
+
+        :param name: A string with range value in A1 notation, e.g. 'A1:A5'.
+
+        Alternatively, you may specify numeric boundaries. All values
+        index from 1 (one):
+
+        :param first_row: Integer row number
+        :param first_col: Integer row number
+        :param last_row: Integer row number
+        :param last_col: Integer row number
+
+        .. versionadded:: 3.4
+
+        """
+        rng = {
+            'sheetId': self.id,
+        }
+
+        if name is not None:
+            start, end = name.split(':')
+            (row_offset, column_offset) = a1_to_rowcol(start)
+            (last_row, last_column) = a1_to_rowcol(end)
+            rng['startRowIndex'] = row_offset - 1
+            rng['endRowIndex'] = last_row
+            rng['startColumnIndex'] = column_offset - 1
+            rng['endColumnIndex'] = last_column
+
+        filter_settings = {
+            'range': rng
+        }
+
+        body = {
+            'requests': [{
+                'setBasicFilter': {
+                    'filter': filter_settings
+                }
+            }]
+        }
+
+        return self.spreadsheet.batch_update(body)
+
+    def clear_basic_filter(self):
+        """Remove the basic filter from a worksheet.
+
+        .. versionadded:: 3.4
+
+        """
+        body = {
+            'requests': [{
+                'clearBasicFilter': {
+                    'sheetId': self.id,
+                }
+            }]
+        }
+
+        return self.spreadsheet.batch_update(body)
+
     def export(self, format):
         """.. deprecated:: 2.0
 
@@ -1451,6 +1774,54 @@ class Worksheet(object):
             new_sheet_name
         )
 
+    @cast_to_a1_notation
+    def merge_cells(self, name, merge_type="MERGE_ALL"):
+        """
+        Merge cells. There are 3 merge types: MERGE_ALL, MERGE_COLUMNS,
+        and MERGE_ROWS.
+
+        :param name: A string with range value in A1 notation, e.g. 'A1:A5'.
+        :param merge_type: (optional) one of MERGE_ALL, MERGE_COLUMNS,
+                           or MERGE_ROWS. Default MERGE_ROWS
+                           See `MergeType`_ in the Sheets API reference.
+
+        Alternatively, you may specify numeric boundaries. All values
+        index from 1 (one):
+
+        :param first_row: Integer row number
+        :param first_col: Integer row number
+        :param last_row: Integer row number
+        :param last_col: Integer row number
+
+        :return the response body from the request
+
+        .. MergeType: https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#MergeType
+
+        """
+
+        start, end = name.split(':')
+        (row_offset, column_offset) = a1_to_rowcol(start)
+        (last_row, last_column) = a1_to_rowcol(end)
+
+        body = {
+            "requests": [
+                {
+                    "mergeCells": {
+                        "mergeType": merge_type,
+                        "range": {
+                            "sheetId": self.id,
+                            "startRowIndex": row_offset - 1,
+                            "endRowIndex": last_row,
+                            "startColumnIndex": column_offset - 1,
+                            "endColumnIndex": last_column,
+                        }
+                    }
+                }
+            ]
+        }
+
+        return self.spreadsheet.batch_update(body)
+
 
 class Cell(object):
     """An instance of this class represents a single cell
